"""module with trailing underscores everywhere"""


class class_:
    """Class"""

    def method_(_arg):  # NoQA: N805
        """Method"""
        pass


def function_(_arg):
    """Function"""
    pass
